# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Read inline dat strings."""

from functools import partial
from typing import Any

from fourcipp.utils.metadata import Primitive
from fourcipp.utils.type_hinting import (
    Extractor,
    LineCastingDict,
    LineListExtractor,
    NestedCastingDict,
    T,
)

# Metadata types currently supported
SUPPORTED_METADATA_TYPES = Primitive.PRIMITIVE_TYPES + ["enum", "vector"]


def to_dat_string(object: Any) -> str:
    """Convert object to dat style string.
    Args:
        data: Object to be casted

    Returns:
        Object as dict
    """
    if isinstance(object, list):
        return " ".join([str(d) for d in object])
    elif isinstance(object, bool):
        return str(object).lower()
    return str(object)


def _left_pop(line_list: list[str], n_entries: int) -> list[str]:
    """Pop entries the beginning of a list.

    Args:
        line_list: List to extract the entries
        n_entries: Number of entries starting from the beginning of the list

    Returns:
        Extracted entries
    """
    entries = line_list[:n_entries]
    del line_list[:n_entries]
    return entries


def _extract_entry(line_list: list[str], extractor: Extractor[T]) -> T:
    """Extract a single entry from a line list.

    Args:
        line_list: List to extract the entries
        extractor: Function to cast the string into the desired object

    Returns:
        Casted object
    """
    return extractor(_left_pop(line_list, 1)[0])


def _extract_vector(
    line_list: list[str], extractor: Extractor[T], size: int
) -> list[T]:
    """Extract a vector entry from a line list.

    Args:
        line_list: List to extract the entries
        extractor: Function to cast the string into the desired object
        size: Vector size

    Returns:
        Casted vector object
    """
    return [extractor(e) for e in _left_pop(line_list, size)]


def _extract_enum(line_list: list[str], choices: list[str]) -> str:
    """Extract enum entry from a line list.

    Args:
        line_list: List to extract the entries
        choices: Choices for the enum

    Returns:
        Valid enum entry
    """
    entry = _left_pop(line_list, 1)[0]
    if not entry in choices:
        raise ValueError(f"Unknown entry {entry}, valid choices are {choices}")
    return entry


def _entry_casting_factory(spec: dict) -> LineListExtractor:
    """Create the casting function for a spec.

    Args:
        spec: 4C metadata style object description

    Returns:
        Casting function for the spec
    """

    primitive_extractors = Primitive.PRIMITIVE_TYPES_TO_PYTHON.copy()
    primitive_extractors["bool"] = lambda v: {
        "true": True,
        "yes": True,
        "1": True,
        "false": False,
        "no": False,
        "0": False,
    }[v]

    if spec["type"] in primitive_extractors:
        extractor = primitive_extractors[spec["type"]]
        return partial(_extract_entry, extractor=extractor)
    elif spec["type"] == "vector":
        value_type = primitive_extractors[spec["value_type"]["type"]]
        return partial(_extract_vector, extractor=value_type, size=spec["size"])
    elif spec["type"] == "enum":
        choices = [s["name"] for s in spec["choices"]]
        return partial(_extract_enum, choices=choices)
    else:
        raise NotImplementedError(f"Entry type {spec['type']} not supported.")


def casting_factory(fourc_metadata: dict) -> LineCastingDict:
    """Create casting object for the specs.

    Args:
        fourc_metadata: 4C metadata style object description

    Returns:
        Casting object for the specs by name
    """
    metadata_type = fourc_metadata["type"]

    if metadata_type == "all_of":
        specs: LineCastingDict = {}

        for spec_i in fourc_metadata["specs"]:
            if spec_i["type"] in SUPPORTED_METADATA_TYPES:
                spec_name: str = spec_i["name"]
                specs[spec_name] = _entry_casting_factory(spec_i)
            else:
                raise NotImplementedError(f"Entry type {spec_i['type']} not supported.")

        return specs

    raise NotImplementedError(f"First entry has to be an all_of")


def nested_casting_factory(fourc_metadata: dict) -> NestedCastingDict:
    """Create nested casting object for the specs.

    Args:
        fourc_metadata: 4C metadata style object description

    Returns:
        Casting object for the specs by name
    """

    if fourc_metadata["type"] in SUPPORTED_METADATA_TYPES:
        type_name: str = fourc_metadata["name"]
        return {type_name: _entry_casting_factory(fourc_metadata)}

    # Supported collections
    if fourc_metadata["type"] in ["all_of", "group", "one_of"]:
        specs: NestedCastingDict = {}
        for spec_i in fourc_metadata["specs"]:
            specs.update(nested_casting_factory(spec_i))

        if fourc_metadata["type"] == "group":
            type_name = fourc_metadata["name"]
            return {type_name: specs}  # type: ignore[dict-item]
        else:
            return specs
    else:
        raise NotImplementedError(f"Entry type {fourc_metadata['type']} not supported.")


def inline_dat_read(line_list: list, keyword_casting: LineCastingDict) -> dict:
    """Read inline dat to dict.

    Note: This function is not able to read nested containers such as groups. This would diminish
    performance and is not needed for the legacy sections.

    Args:
        line_list: List to extract the entries
        keyword_casting: Dict with the casting

    Returns:
        Entry as dict
    """
    entry: dict = {}
    while line_list:
        key = line_list.pop(0)
        # Raises Error if an entry was provided twice
        if key in entry:
            raise KeyError(f"The entry {key} was provided already: {entry}")
        entry[key] = keyword_casting[key](line_list)

    return entry



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Element io."""

from typing import TypeAlias

from fourcipp import CONFIG
from fourcipp.legacy_io.inline_dat import (
    _extract_vector,
    inline_dat_read,
    nested_casting_factory,
    to_dat_string,
)
from fourcipp.utils.type_hinting import LineCastingDict

ElementCastingDict: TypeAlias = dict[str, dict[str, LineCastingDict]]


def element_data_casting_factory(
    legacy_element_specs: dict,
) -> ElementCastingDict:
    """Create element data casting dict.

    Args:
        legacy_element_specs: Element data

    Returns:
        element casting dict
    """
    elements_data_casting_dict: ElementCastingDict = {}
    for element_type, element_data in legacy_element_specs.items():
        element_type_data: dict[str, LineCastingDict] = {}

        for cell in element_data:
            element_type_data[cell["cell_type"]] = nested_casting_factory(cell["spec"])  # type: ignore[assignment]

        elements_data_casting_dict[element_type] = element_type_data

    return elements_data_casting_dict


_element_data_casting: ElementCastingDict = element_data_casting_factory(
    CONFIG.fourc_metadata["legacy_element_specs"]
)


CELL_TYPES = CONFIG.fourc_metadata["cell_types"]


def read_element(
    line: str, element_data_casting: ElementCastingDict = _element_data_casting
) -> dict:
    """Read a element line.

    Args:
        line: Inline dat description of the element
        elements_casting: Element casting dict.

    Returns:
        element as dict
    """
    line_list = line.split()

    # First entry is always the element id starting from 1
    element_id = int(line_list.pop(0))

    # Second entry is always the element type
    element_type = line_list.pop(0)

    # Third entry is the cell type
    cell_type = line_list.pop(0)

    # The next entries are the nodes of the element
    connectivity = _extract_vector(
        line_list, int, CELL_TYPES[cell_type]["number_of_nodes"]
    )

    # Element
    element = {
        "id": element_id,
        "cell": {
            "type": cell_type,
            "connectivity": connectivity,
        },
    }

    # Element data
    element["data"] = {"type": element_type} | inline_dat_read(
        line_list, element_data_casting[element_type][cell_type]
    )

    return element


def write_element(element: dict) -> str:
    """Write element as inline dat style.

    Args:
        element: Element description

    Returns:
        element line
    """
    line = " ".join(
        [
            to_dat_string(element["id"]),
            to_dat_string(element["data"]["type"]),
            to_dat_string(element["cell"]["type"]),
            to_dat_string(element["cell"]["connectivity"]),
        ]
    )
    for k, v in element["data"].items():
        if k == "type":
            continue
        line += " " + k + " " + to_dat_string(v)
    return line



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Particle io.

Once this section is implemented in 4C using InputSpec, this file can be
simplified.
"""

from fourcipp import CONFIG
from fourcipp.legacy_io.inline_dat import (
    casting_factory,
    inline_dat_read,
    to_dat_string,
)
from fourcipp.utils.type_hinting import LineCastingDict

_particle_casting = casting_factory(CONFIG.fourc_metadata["legacy_particle_specs"])


def read_particle(
    line: str, particle_casting: LineCastingDict = _particle_casting
) -> dict:
    """Read particle.

    Args:
        line: Inline dat description of the particle
        particle_casting: Particle casting dict.

    Returns:
        Particle section as dict
    """
    return inline_dat_read(line.split(), particle_casting)


def write_particle(particle: dict) -> str:
    """Write particles section.

    Args:
        particle: Particle as dict

    Returns:
        Particle line
    """
    line = ""

    for k, v in particle.items():
        if line:
            line += " "

        line += k + " " + to_dat_string(v)

    return line



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test read me example."""

import pathlib
import subprocess

import pytest

from fourcipp.fourc_input import FourCInput


@pytest.fixture(name="readme_example")
def fixture_readme_example(tmp_path):
    """Extract readme example and create file."""

    input_file_path = tmp_path / "other_input.4C.yaml"
    fourc_input = FourCInput()
    fourc_input["TITLE"] = {"some": "thing"}
    fourc_input.dump(input_file_path)

    readme = pathlib.Path(__file__).parents[2] / "README.md"
    example = readme.read_text(encoding="utf-8")
    example_marker = "<!--example, do not remove this comment-->"
    example = example.split(example_marker)[1]

    # Add missing path
    escaped_path = repr(str(input_file_path))  # For windows compatibility
    example = [f"input_file_path = {escaped_path}"] + [
        line for line in example.split("\n") if not "```" in line
    ]
    example = "\n".join(example)
    example_path = tmp_path / "example.py"

    # Create file
    example_path.write_text(example)

    return example_path


def test_readme_example(readme_example, tmp_path):
    """Test if the example runs."""
    command = f"python {readme_example} > {tmp_path / 'output.log'} 2>&1"
    return_code = subprocess.call(command, shell=True)  # nosec

    # Exit code -> script failed
    if return_code:
        raise Exception(
            f"Example run failed: {command}\n\nOutput: {(tmp_path / 'output.log').read_text(encoding='utf-8')}"
        )



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test node topology io."""

import pytest

from fourcipp.legacy_io.node_topology import (
    read_node_topology,
    write_node_topology,
)


@pytest.mark.parametrize(
    "line",
    [
        "NODE 9 DNODE 1",
        "NODE 9 DLINE 1",
        "NODE 9 DSURFACE 1",
        "NODE 9 DVOL 1",
    ],
)
def test_dtopology_read_and_write(line):
    """Test DNODE, DLINE, DSURFACE, DVOL read and write."""
    assert line.split() == write_node_topology(read_node_topology(line)).split()


@pytest.mark.parametrize(
    "line",
    [
        "CORNER fluid x- y- z+ DNODE 1",
        "EDGE fluid y+ y- DLINE 1",
        "SIDE fluid y+ DSURFACE 1",
        "VOLUME fluid DVOL 1",
    ],
)
def test_domain_read_and_write(line):
    """Test CORNER, EDGE, SIDE, VOLUME read and write."""
    assert line.split() == write_node_topology(read_node_topology(line)).split()


def test_read_node_topology_error():
    """Test node topology error."""
    unknown_node = "NOPE something"
    with pytest.raises(ValueError, match="Unknown type"):
        read_node_topology(unknown_node)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test node reader and writer."""

import pytest

from fourcipp.legacy_io.node import read_node, write_node


@pytest.mark.parametrize(
    "node_line",
    [
        "NODE 5 COORD 0.0 0.1 0.2",
        "CP 5 COORD 0.0 0.1 0.2 0.3",
        "FNODE 5 COORD 0.0 0.1 0.2 CIR 1.1 1.2 1.3 TAN 1.1 1.2 1.3 RAD 1.1 1.2 1.3 HELIX 1.4 TRANS 1.4 FIBER1 1.1 1.2 1.3 FIBER2 1.1 1.2 1.3",
    ],
)
def test_node_read_and_write(node_line):
    """Test nodes read and write."""
    assert node_line.split() == write_node(read_node(node_line)).split()


def test_unknown_node_read():
    """Test unknown node read."""
    unknown_node = "NOPE 5 COORD 0.0 0.1 0.2"
    with pytest.raises(ValueError, match="Unknown node type"):
        read_node(unknown_node)


def test_unknown_node_write():
    """Test unknown node write."""
    unknown_node = {"id": 5, "COORD": [0.0, 0.1, 0.2], "data": {"type": "NOPE"}}
    with pytest.raises(ValueError, match="Unknown node type"):
        write_node(unknown_node)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""YAML io."""

import json
import pathlib
from typing import Callable

import regex
import ryml

from fourcipp.utils.type_hinting import Path


def load_yaml(path_to_yaml_file: Path) -> dict:
    """Load yaml files.

    rapidyaml is the fastest yaml parsing library we could find. Since it returns custom objects we
    use the library to emit the objects to json and subsequently read it in using the json library.
    This is still two orders of magnitude faster compared to other yaml libraries.

    Args:
        path_to_yaml_file: Path to yaml file

    Returns:
       Loaded data
    """

    json_str = ryml.emit_json(
        ryml.parse_in_arena(pathlib.Path(path_to_yaml_file).read_bytes())
    )

    # Convert `inf` to a string to avoid JSON parsing errors, see https://github.com/biojppm/rapidyaml/issues/312
    json_str = regex.sub(r":\s*(-?)inf\b", r': "\1inf"', json_str)

    # Convert floats that are missing digits on either side of the decimal point
    # so .5 to 0.5 and 5. to 5.0
    json_str = regex.sub(r":\s*(-?)\.([0-9]+)", r": \g<1>0.\2", json_str)
    json_str = regex.sub(r":\s*(-?)([0-9]+)\.(\D)", r": \1\2.0\3", json_str)

    data = json.loads(json_str)

    return data


def dict_to_yaml_string(
    data: dict,
    sort_function: Callable[[dict], dict] | None = None,
    use_fourcipp_yaml_style: bool = True,
) -> str:
    """Dump dict as yaml.

    The FourCIPP yaml style sets flow
    Args:
        data: Data to dump.
        sort_function: Function to sort the data.
        use_fourcipp_yaml_style: If FourCIPP yaml style is to be used

    Returns:
        YAML string representation of the data
    """

    if sort_function is not None:
        data = sort_function(data)

    # Convert dictionary into a ryml tree
    tree = ryml.parse_in_arena(bytearray(json.dumps(data).encode("utf8")))

    def check_is_vector(tree: ryml.Tree, node_id: int) -> bool:
        """Check if sequence is of ints, floats or sequence there of.

        In 4C metadata terms, list of strings, bools, etc. could also be vectors.
        For the sake of simplicity these are omitted.

        Args:
            tree (ryml.Tree): Tree to check
            node_id (int): Node id

        Returns:
            returns if entry is a vector
        """

        for sub_node, _ in ryml.walk(tree, node_id):
            # Ignore the root node
            if sub_node == node_id:
                continue

            # If sequence contains a dict
            if tree.is_map(sub_node):
                return False

            # If sequence contains a sequence
            elif tree.is_seq(sub_node):
                if not check_is_vector(tree, sub_node):
                    return False

            # Else it's a value
            else:
                val = tree.val(sub_node).tobytes().decode("ascii")
                is_not_numeric = (
                    tree.is_val_quoted(sub_node)  # string
                    or tree.val_is_null(sub_node)  # null
                    or val == "true"  # bool
                    or val == "false"  # bool
                )
                if is_not_numeric:
                    return False

        return True

    # Change style bits to avoid JSON output and format vectors correctly
    # see https://github.com/biojppm/rapidyaml/issues/520
    for node_id, depth in ryml.walk(tree):
        if tree.is_map(node_id):
            tree.set_container_style(node_id, ryml.NOTYPE)
        elif tree.is_seq(node_id):
            if (
                not use_fourcipp_yaml_style  # do not do special formatting
                or depth == 1  # is a section
                or not check_is_vector(tree, node_id)  # is not a vector
            ):
                tree.set_container_style(node_id, ryml.NOTYPE)

        if tree.has_key(node_id):
            tree.set_key_style(node_id, ryml.NOTYPE)

    yaml_string = ryml.emit_yaml(tree)

    if use_fourcipp_yaml_style:
        # add spaces after commas in vectors
        yaml_string = regex.sub(r"(?<=\d),(?=\d)|(?<=\]),(?=\[)", ", ", yaml_string)

    return yaml_string


def dump_yaml(
    data: dict,
    path_to_yaml_file: Path,
    sort_function: Callable[[dict], dict] | None = None,
    use_fourcipp_yaml_style: bool = True,
) -> None:
    """Dump yaml to file.

    Args:
        data: Data to dump.
        path_to_yaml_file: Yaml file path
        sort_function: Function to sort the data
        use_fourcipp_yaml_style: If FourCIPP yaml style is to be used
    """
    pathlib.Path(path_to_yaml_file).write_text(
        dict_to_yaml_string(data, sort_function, use_fourcipp_yaml_style),
        encoding="utf-8",
    )



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Not set utils."""


class NotSet:
    """Not set object."""

    def __init__(self, expected: object = object) -> None:
        """Not set object.

        Args:
            expected: Expected object to to display
        """
        self.expected = expected

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"NotSet({self.expected})"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"NotSet({self.expected})"


NOT_SET = NotSet()


def check_if_set(obj: object) -> bool:
    """Check if object or is NotSet.

    Args:
        obj: Object to check

    Returns:
        True if object is set
    """
    # Check if object is not of type _NotSet, i.e. it has a value
    return not isinstance(obj, NotSet)


def pop_arguments(key: str, default: object = NOT_SET) -> tuple:
    """Create arguments for the pop method.

    We need this utility since pop is not implemented using kwargs, instead the default is checked
     via the number of arguments.

    Args:
        key: Key to pop the value for
        default: Default value to return in case of the pop value.

    Returns:
        Arguments for pop
    """
    if check_if_set(default):
        return (key, default)
    else:
        return (key,)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test dict utils."""

import numpy as np
import pytest

from fourcipp.utils.dict_utils import (
    _get_dict,
    change_default,
    compare_nested_dicts_or_lists,
    get_entry,
    make_default_explicit,
    make_default_implicit,
    remove,
    rename_parameter,
    replace_value,
    sort_alphabetically,
    sort_by_key_order,
)


@pytest.mark.parametrize(
    "obj,reference_obj",
    [
        ([1, 2, 3], [1, 2, 3]),
        ({"a": 1, "b": 2, "c": 3}, {"c": 3, "b": 2, "a": 1}),
        (
            {
                "a": 1,
                "b": {"g": [{"a": 2, "d": [{"f": 2}, {"e": 2.1}]}], "a": 2},
                "c": [3, True],
            },
            {
                "c": [3, True],
                "b": {"a": 2, "g": [{"a": 2, "d": [{"f": 2}, {"e": 2.1}]}]},
                "a": 1,
            },
        ),
        ("text", "text"),
    ],
)
def test_compare(obj, reference_obj):
    """Test comparison."""
    assert compare_nested_dicts_or_lists(obj, reference_obj)


@pytest.mark.parametrize(
    "obj,reference_obj",
    [
        ([1, 2, 3], [1, 2.1, 3]),
        ({"a": 1, "b": 2, "d": 3}, {"c": 3, "b": 2, "a": 1}),
        (
            {
                "a": 1,
                "b": {"g": [{"a": 2, "d": [{"f": 2}, {"e": 2.1}]}], "a": 2},
                "c": [3.1, True],
            },
            {
                "c": [3, True],
                "b": {"a": 2, "g": [{"a": 2, "d": [{"f": 2}, {"e": 2.1}]}]},
                "a": 1,
            },
        ),
        (True, False),
        (True, "text"),
        ("some", "text"),
        ([1, 2], [1, 2, 3]),
        (True, 1.0),  # Important test since isinstance(True, int) is true
    ],
)
def test_compare_failure(obj, reference_obj):
    """Test comparison failure."""
    with pytest.raises(AssertionError):
        compare_nested_dicts_or_lists(obj, reference_obj)


@pytest.mark.parametrize(
    "obj,reference_obj",
    [
        (0, -0.00000001),
        ([1, 2, 3], [1, 2.0, 3]),
        ({"a": 1, "b": 2, "c": 3}, {"c": 3, "b": 2, "a": 1.000005}),
        (
            {
                "a": 1,
                "b": {"g": [{"a": 2, "d": [{"f": 2.0000005}, {"e": 2.1}]}], "a": 2},
                "c": [3, True],
            },
            {
                "c": [3, True],
                "b": {"a": 2, "g": [{"a": 2, "d": [{"f": 2.0}, {"e": 2.1}]}]},
                "a": 1,
            },
        ),
    ],
)
def test_compare_allow_int_as_float(obj, reference_obj):
    """Test comparison but compare ints to floats."""
    assert compare_nested_dicts_or_lists(
        obj, reference_obj, allow_int_vs_float_comparison=True
    )


@pytest.mark.parametrize(
    "obj,reference_obj",
    [
        (0, -0.001),
        ([1, 2, 3], [1, 2.02, 3]),
        ({"a": 1, "b": 2, "c": 3}, {"c": 3, "b": 2, "a": 1.005}),
        (
            {
                "a": 1,
                "b": {"g": [{"a": 2, "d": [{"f": 2.05}, {"e": 2.1}]}], "a": 2},
                "c": [3, True],
            },
            {
                "c": [3, True],
                "b": {"a": 2, "g": [{"a": 2, "d": [{"f": 2.0}, {"e": 2.1}]}]},
                "a": 1,
            },
        ),
    ],
)
def test_compare_allow_int_as_float_failure(obj, reference_obj):
    """Test comparison failure due to tolerance."""
    with pytest.raises(AssertionError, match="The numerics are not close"):
        compare_nested_dicts_or_lists(
            obj, reference_obj, allow_int_vs_float_comparison=True
        )


def custom_compare(obj, reference_obj):
    """Examplary custom comparison function.

    Args:
        obj (object): Object for comparison
        reference_obj (object): Reference object

    Returns:
        bool: True if objects are equal
    """
    # Handle np.ndarray
    if isinstance(obj, np.ndarray) and isinstance(reference_obj, np.ndarray):
        if not np.array_equal(obj, reference_obj):
            raise AssertionError("Custom compare")
        return True

    # Nothing to return if objects are not compared here


def test_compare_with_custom_function():
    """Test comparison with custom compare."""
    obj = {"b": [{"a": np.ones(2)}], "c": {"a": 1}}
    reference_obj = {"b": [{"a": np.ones(2)}], "c": {"a": 1}}

    assert compare_nested_dicts_or_lists(
        obj, reference_obj, custom_compare=custom_compare
    )


def test_compare_with_custom_function_failure():
    """Test comparison failure with custom compare."""
    obj = {"b": [{"a": np.ones(2)}], "c": {"a": 1}}
    reference_obj = {"b": [{"a": 2 * np.ones(2)}], "c": {"a": 1}}

    with pytest.raises(AssertionError, match="Custom compare"):
        compare_nested_dicts_or_lists(obj, reference_obj, custom_compare=custom_compare)


@pytest.fixture(name="nested_input_dict")
def fixture_nested_input_dict():
    """Nested dict feature."""
    nested_input_dict = {
        "a": {
            "b": [
                {
                    "c": 1,
                    "d": {"e": {"g": [5, 1]}, "h": [{"a": 5}, {"b": 5}], "f": 2},
                    "i": 4,
                },
                {
                    "c": 2,
                    "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                    "i": 5,
                },
                {"c": 1, "d": {"e": {"g": [7, 3]}, "h": [{"a": 7}, {"b": 1}], "f": 2}},
            ],
            "f": 3,
        }
    }
    return nested_input_dict


@pytest.mark.parametrize(
    "keys,value",
    [
        (
            ("a"),
            [
                {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                }
            ],
        ),
        (
            ("a", "b"),
            [
                [
                    {
                        "c": 1,
                        "d": {"e": {"g": [5, 1]}, "h": [{"a": 5}, {"b": 5}], "f": 2},
                        "i": 4,
                    },
                    {
                        "c": 2,
                        "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                        "i": 5,
                    },
                    {
                        "c": 1,
                        "d": {"e": {"g": [7, 3]}, "h": [{"a": 7}, {"b": 1}], "f": 2},
                    },
                ]
            ],
        ),
        (("a", "b", "c"), [1, 2, 1]),
        (("a", "b", "d", "h", "b"), [5, 1]),
    ],
)
def test_get_entry(nested_input_dict, keys, value):
    """Test get entry."""
    data = get_entry(nested_input_dict, keys)
    for i, entry in enumerate(data):
        assert entry == value[i]


@pytest.mark.parametrize(
    "keys",
    [
        ("not existing"),  # first not exists
        ("a", "b", "d", "h", "b"),  # does not exist but not in every entry of the list
        ("a", "not existing"),  # does not exist in the last key
    ],
)
def test_get_entry_not_optional(nested_input_dict, keys):
    """Test get entry without optional setting."""
    with pytest.raises(KeyError):
        for _ in get_entry(nested_input_dict, keys, optional=False):
            pass


@pytest.mark.parametrize(
    "keys, value",
    [
        (("a"), {}),
        (("a", "b"), {"a": {"f": 3}}),
        (
            ("a", "b", "d", "h", "b"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {"e": {"g": [7, 3]}, "h": [{"a": 7}, {}], "f": 2},
                        },
                    ],
                    "f": 3,
                }
            },
        ),
    ],
)
def test_remove(nested_input_dict, keys, value):
    """Test remove parameter."""
    remove(nested_input_dict, keys)
    assert nested_input_dict == value


@pytest.mark.parametrize(
    "keys, value",
    [
        (("a"), {"a": "new_value"}),
        (("a", "b"), {"a": {"f": 3, "b": "new_value"}}),
        (
            ("a", "b", "d", "h", "b"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": "new_value"}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": "new_value"}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                }
            },
        ),
    ],
)
def test_replace_value(nested_input_dict, keys, value):
    """Test replace value."""
    replace_value(nested_input_dict, keys, "new_value")
    assert nested_input_dict == value


@pytest.mark.parametrize(
    "keys, value",
    [
        (
            ("a", "j"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                    "j": "default",
                }
            },
        ),
        (
            ("a", "b", "i"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                            "i": "default",
                        },
                    ],
                    "f": 3,
                }
            },
        ),
    ],
)
def test_make_default_explicit(nested_input_dict, keys, value):
    """Test make default explicit."""
    make_default_explicit(nested_input_dict, keys, "default")
    assert nested_input_dict == value


@pytest.mark.parametrize(
    "keys, value, default",
    [
        (
            ("a", "f"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                }
            },
            3,
        ),
        (
            ("a", "b", "i"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                },
            },
            5,
        ),
    ],
)
def test_make_default_implicit(nested_input_dict, keys, value, default):
    """Test make default implicit."""
    make_default_implicit(nested_input_dict, keys, default)
    assert nested_input_dict == value


@pytest.mark.parametrize(
    "keys, value, old_default, new_default",
    [
        (
            ("a", "b", "i"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                            "i": 3,
                        },
                    ],
                    "f": 3,
                }
            },
            3,
            5,
        )
    ],
)
def test_make_change_default(nested_input_dict, keys, value, old_default, new_default):
    """Test change default."""
    change_default(nested_input_dict, keys, old_default, new_default)
    assert nested_input_dict == value


@pytest.mark.parametrize(
    "keys, value",
    [
        (
            ("a", "b"),
            {
                "a": {
                    "new_name": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "i": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "i": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                }
            },
        ),
        (
            ("a", "b", "i"),
            {
                "a": {
                    "b": [
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [5, 1]},
                                "h": [{"a": 5}, {"b": 5}],
                                "f": 2,
                            },
                            "new_name": 4,
                        },
                        {
                            "c": 2,
                            "d": {"e": {"g": [6, 2]}, "h": [{"a": 6}], "f": 2},
                            "new_name": 5,
                        },
                        {
                            "c": 1,
                            "d": {
                                "e": {"g": [7, 3]},
                                "h": [{"a": 7}, {"b": 1}],
                                "f": 2,
                            },
                        },
                    ],
                    "f": 3,
                }
            },
        ),
    ],
)
def test_rename_parameter(nested_input_dict, keys, value):
    """Test renaming parameter."""
    rename_parameter(nested_input_dict, keys, "new_name")
    assert nested_input_dict == value


@pytest.mark.parametrize("nested_dict", [{"a": "string"}, {"a": ["string"]}])
def test_get_dict_failure(nested_dict):
    """Test _get_dict failure."""
    with pytest.raises(TypeError):
        for _ in _get_dict(nested_dict, ("a")):
            pass


def test_get_dict_optional(nested_input_dict):
    """Test _get_dict one of the keys does not exist."""
    result = "some value"
    for entry in _get_dict(nested_input_dict, ("a", "not_existing", "b")):
        result = entry
    # assert if nothing changed
    assert result == "some value"


def test_sort_by_key_order_basic():
    """Test sorting by key order."""

    data = {"b": 2, "a": 1, "c": 3}

    assert sort_by_key_order(data, ["a", "b", "c"]) == {"a": 1, "b": 2, "c": 3}


def test_sort_by_key_order_error_mismatched_keys():
    """Test sorting by key order with mismatched keys."""

    data = {"b": 2, "a": 1, "c": 3}

    with pytest.raises(
        ValueError, match="'key_order' must include all keys in the dictionary!"
    ):
        sort_by_key_order(data, ["a", "b"])


def test_sort_alphabetically():
    """Test alphabetical sorting."""

    data = {"A": 1, "b": 2, "a": 0, "B": 3, "c": 4, "C": 5}

    assert sort_alphabetically(data) == {"a": 0, "A": 1, "b": 2, "B": 3, "c": 4, "C": 5}



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Modules related to legacy io."""

from collections.abc import Callable, Sequence

from fourcipp.legacy_io.element import read_element, write_element
from fourcipp.legacy_io.node import read_node, write_node
from fourcipp.legacy_io.node_topology import read_node_topology, write_node_topology
from fourcipp.legacy_io.particle import read_particle, write_particle
from fourcipp.utils.type_hinting import T


def _iterate_and_evaluate(function: Callable[..., T], iterable: Sequence) -> list[T]:
    """Iterate through lines and evaluate function on them.

    Args:
        function: Function to be called
        iterable: List of data

    Returns:
        List evaluate as desired
    """
    return [function(line) for line in iterable]


def interpret_legacy_section(
    legacy_section: str, section_data: list, known_legacy_sections: list
) -> dict | list:
    """Interpret legacy section.

    Transform line into usable data.

    Args:
        legacy_section: Section name
        section_data: Section data
        known_legacy_sections: Known legacy sections

    Returns:
        Interpreted data
    """
    if legacy_section not in known_legacy_sections:
        raise ValueError(
            f"Section {legacy_section} is not a known legacy section. Current legacy sections are {', '.join(known_legacy_sections)}"
        )

    match legacy_section:
        case "PARTICLES":
            return _iterate_and_evaluate(read_particle, section_data)

        case "NODE COORDS":
            return _iterate_and_evaluate(read_node, section_data)

        case _ if legacy_section.endswith("ELEMENTS"):
            return _iterate_and_evaluate(read_element, section_data)

        case _ if legacy_section.endswith("NODE TOPOLOGY"):
            return _iterate_and_evaluate(read_node_topology, section_data)

        case _:
            raise NotImplementedError(
                f"Legacy section {legacy_section} is not implemented."
            )


def interpret_legacy_sections(
    legacy_sections: dict, known_legacy_sections: list
) -> dict:
    """Interpret all legacy sections.

    Args:
        legacy_sections: Legacy sections and data
        known_legacy_sections: Known legacy sections

    Returns:
        Interpreted legacy sections
    """
    for legacy_section, section in legacy_sections.items():
        legacy_sections[legacy_section] = interpret_legacy_section(
            legacy_section, section, known_legacy_sections
        )
    return legacy_sections


def inline_legacy_section(
    legacy_section: str, section_data: dict | list, known_legacy_sections: list
) -> list:
    """Inline legacy section.

    Transform dict form of the section into the inline dat style.

    Args:
        legacy_section: Section name
        section_data: Section data
        known_legacy_sections: Known legacy sections

    Returns:
        Inlined data
    """

    match legacy_section:
        case "PARTICLES":
            if not isinstance(section_data, Sequence):
                raise TypeError("Expected the section data to be of type list.")
            return _iterate_and_evaluate(write_particle, section_data)

        case "NODE COORDS":
            if not isinstance(section_data, Sequence):
                raise TypeError("Expected the section data to be of type list.")
            return _iterate_and_evaluate(write_node, section_data)

        case _ if legacy_section.endswith("ELEMENTS"):
            if not isinstance(section_data, Sequence):
                raise TypeError("Expected the section data to be of type list.")
            return _iterate_and_evaluate(write_element, section_data)

        case _ if legacy_section.endswith("NODE TOPOLOGY"):
            if not isinstance(section_data, Sequence):
                raise TypeError("Expected the section data to be of type list.")
            return _iterate_and_evaluate(write_node_topology, section_data)

        case _:
            raise ValueError(
                f"Section {legacy_section} is not a known legacy section. Current legacy sections are {', '.join(known_legacy_sections)}"
            )


def inline_legacy_sections(legacy_sections: dict, known_legacy_sections: list) -> dict:
    """Inline all legacy sections.

    Args:
        legacy_sections: Legacy sections and data
        known_legacy_sections: Known legacy sections

    Returns:
        Inline legacy sections
    """

    for legacy_section, section in legacy_sections.items():
        legacy_sections[legacy_section] = inline_legacy_section(
            legacy_section, section, known_legacy_sections
        )
    return legacy_sections



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Node io.

Once this section is implemented in 4C using InputSpec, this file can be
simplified.
"""

from collections.abc import Callable
from functools import partial
from typing import Any, Literal

from fourcipp.legacy_io.inline_dat import _extract_entry, _extract_vector, to_dat_string

_FNODE_CASTING: dict[str, Callable] = {
    "CIR": partial(_extract_vector, extractor=float, size=3),
    "TAN": partial(_extract_vector, extractor=float, size=3),
    "RAD": partial(_extract_vector, extractor=float, size=3),
    "HELIX": partial(_extract_entry, extractor=float),
    "TRANS": partial(_extract_entry, extractor=float),
}


def read_node(line: str) -> dict:
    """Read node.

    Args:
        line: Inline dat description of the element

    Returns:
        Node as dict
    """
    line_list = line.split()

    # First entry is the node type
    node_type = line_list.pop(0)

    # Second entry is the node id
    node_id = int(line_list.pop(0))

    # Read the coords
    line_list.pop(0)
    coordinate = _extract_vector(line_list, extractor=float, size=3)

    node: dict[Literal["id", "COORD", "data"], dict | Any] = {
        "id": node_id,
        "COORD": coordinate,
        "data": {"type": node_type},
    }

    if node_type == "NODE":
        return node

    if node_type == "CP":
        node["data"]["weight"] = float(line_list[0])
        return node

    if node_type == "FNODE":
        while line_list:
            key = line_list.pop(0)
            if key.startswith("FIBER"):
                node["data"][key] = _extract_vector(line_list, extractor=float, size=3)
                continue
            node["data"][key] = _FNODE_CASTING[key](line_list)
        return node

    raise ValueError(f"Unknown node type {node_type}")


def write_node(node: dict) -> str:
    """Write node as line.

    Args:
        node: Node as dict

    Returns:
        Node as line
    """
    node_type = node["data"]["type"]
    line = f"{node_type} {node['id']} COORD {to_dat_string(node['COORD'])}"

    if node_type == "NODE":
        return line

    if node_type == "CP":
        return line + " " + str(node["data"]["weight"])

    if node_type == "FNODE":
        for k, v in node["data"].items():
            if k == "type":
                continue
            line += f" {k} {to_dat_string(v)}"
        return line

    raise ValueError(f"Unknown node type {node_type}")



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test element reader and writer."""

import pytest

from fourcipp import CONFIG
from fourcipp.legacy_io.element import (
    read_element,
    write_element,
)

CELL_TYPES = CONFIG.fourc_metadata["cell_types"]
from .utils import reference_value_from_all_of


def inline_element_from_spec(element_type, cell_type, element_data_specs):
    """Generate element line from specs.

    Args:
        element_type (str): Element type
        cell_type (str): Cell type
        element_data_specs (dict): Metadata for all element data and fields

    Returns:
        str: inline element
    """
    cell_nodes = CELL_TYPES[cell_type]["number_of_nodes"]
    connectivity = "42 " * cell_nodes

    # Add additional whitespaces to check the reader
    return (
        f"42 {element_type} {cell_type} {connectivity}"
        + reference_value_from_all_of(element_data_specs)
    )


def generate_elements_from_metadatafile():
    """Generate all possible elements from metadata.

    Returns:
        list: list of inline elements
    """
    data = CONFIG.fourc_metadata["legacy_element_specs"]

    elements = []
    # Loop over element types
    for element_type, element_specs in data.items():
        # Loop over cell types
        for element_data in element_specs:
            cell_type = element_data["cell_type"]
            ele = inline_element_from_spec(
                element_type, cell_type, element_data["spec"]
            )
            elements.append(ele)
    return elements


_REFERENCE_ELEMENTS = generate_elements_from_metadatafile()


@pytest.mark.parametrize("element", _REFERENCE_ELEMENTS)
def test_elements_read_and_write(element):
    """Test elements read and write."""
    element_dict = read_element(element)
    element_line = write_element(element_dict)

    # Split into a list to account for additional whitespaces
    assert element.split() == element_line.split()



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Node topology io.

Once this section is implemented in 4C using InputSpec, this file can be
simplified.
"""

from fourcipp.legacy_io.inline_dat import (
    _extract_entry,
    _extract_enum,
    to_dat_string,
)


def _read_corner(line_list: list) -> dict:
    """Read corner line.

    Args:
        line_list: List to extract the entry

    Returns:
        corner as dict
    """
    corner = {
        "type": "CORNER",
        "discretization_type": line_list.pop(0),
        "corner_description": [
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
        ],
        "d_type": line_list.pop(0),
        "d_id": _extract_entry(line_list, extractor=int),
    }
    return corner


def _read_edge(line_list: list[str]) -> dict:
    """Read edge line.

    Args:
        line_list: List to extract the entry

    Returns:
        edge as dict
    """
    edge = {
        "type": "EDGE",
        "discretization_type": line_list.pop(0),
        "edge_description": [
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
        ],
        "d_type": line_list.pop(0),
        "d_id": _extract_entry(line_list, extractor=int),
    }
    return edge


def _read_side(line_list: list[str]) -> dict:
    """Read side line.

    Args:
        line_list: List to extract the entry

    Returns:
        Side as dict
    """
    side = {
        "type": "SIDE",
        "discretization_type": line_list.pop(0),
        "side_description": [
            _extract_enum(line_list, choices=["x-", "x+", "y-", "y+", "z-", "z+"]),
        ],
        "d_type": line_list.pop(0),
        "d_id": _extract_entry(line_list, extractor=int),
    }
    return side


def _read_volume(line_list: list[str]) -> dict:
    """Read volume line.

    Args:
        line_list: List to extract the entry

    Returns:
        Volume as dict
    """
    volume = {
        "type": "VOLUME",
        "discretization_type": line_list.pop(0),
        "d_type": line_list.pop(0),
        "d_id": _extract_entry(line_list, extractor=int),
    }
    return volume


def _read_domain_topology(line_list: list[str], extractor: str) -> dict:
    """Read domain topology.

    Args:
        line_list: List to extract the entry
        extractor: Type of domain node topology

    Returns:
        Topology entry as a dict
    """
    if extractor == "CORNER":
        return _read_corner(line_list)
    elif extractor == "EDGE":
        return _read_edge(line_list)
    elif extractor == "SIDE":
        return _read_side(line_list)
    elif extractor == "VOLUME":
        return _read_volume(line_list)
    else:
        raise TypeError(f"Unknown entry type {extractor}")


def _read_d_topology(line_list: list[str]) -> dict:
    """Read d topology.
    Args:
        line_list: List to extract the entries

    Returns:
        Topology entry as a dict
    """
    node_id = _extract_entry(line_list, extractor=int)
    d_type = _extract_enum(
        line_list, choices=["DNODE", "DLINE", "DSURFACE", "DSURF", "DVOLUME", "DVOL"]
    )
    d_id = _extract_entry(line_list, extractor=int)

    d_topology = {
        "type": "NODE",
        "node_id": node_id,
        "d_type": d_type,
        "d_id": d_id,
    }
    return d_topology


def read_node_topology(line: str) -> dict:
    """Read topology entry as line.

    Args:
        line: Inline dat description of the topology entry

    Returns:
        Topology entry as a dict
    """
    line_list = line.split()
    extractor = line_list.pop(0)

    if extractor == "NODE":
        return _read_d_topology(line_list)

    if extractor in ["CORNER", "EDGE", "SIDE", "VOLUME"]:
        return _read_domain_topology(line_list, extractor)

    raise ValueError(f"Unknown type {extractor}")


def write_node_topology(topology: dict) -> str:
    """Write topology line.

    Args:
        topology: Topology dict

    Returns:
        Topology entry as line
    """
    return " ".join([to_dat_string(e) for e in topology.values()])
